#==================================================================================================================================================================
# Defines utilities used in other programs
#
# Last updated 16 Jan 2018
#
#
# combinationList(nsel,fromNum)                         Return a maximal list each element of which is a unique selection 
#                                                       containing nsel integers from the integer list covering the range: 1..fromNum.
#
#------------------------------------------------------------------------------------------------------------------------------
# Vector utilities:
#
# includedAngle(vx,vy,vz,wx,wy,wz)                      Given two vectors: v,w  defined by their components in 
#                                                       three orthogonal directions: x,y,z work out the angle between them.
#
# vectorMagnitude(x,y,z)                                The absolute length of a vector.
#
# vectorProduct(x1,y1,z1, x2,y2,z2)                     Return the vector product of two vectors, which can be used
#                                                       to find the direction at right angles to two other vectors.
#
# normaliseVector(x,y,z)                                Returns a unit vector in the same direction as the input vector.
#
#------------------------------------------------------------------------------------------------------------------------------
# Coordinate conversions:
#
# rotateX(x,y,z,a):     Rotate a vector x,y,z around the X axis by right-handed angle a (radians).
# rotateY(x,y,z,a):     Rotate a vector x,y,z around the Y axis by right-handed angle a (radians).
# rotateZ(x,y,z,a):     Rotate a vector x,y,z around the Z axis by right-handed angle a (radians).
#
# ECEF coordinates is a right-handed coordinate system that takes the centre of the Earth as the origin
# - the X axis points towards the 0 of lattitude and longitude on the Earth's surface.
# - the Y axis points towards the 0 of lattitude and 90 deg East.
# - the Z axis is along the Earth's rotation axis pointing North.
#
# gps_to_ecef(latitude,longitude,alt)                   Convert GPS coordinates into Earth Centred, Earth Fixed coordinates.
#
#
# ENU coordinates take the origin as a point on the Earth's surface
# - the X axis points along the meridian of longitude, to the North
# - the Y axis points along a line of lattitude, to the East
# - the Z axis point directly up to the zenith from the origin.
#
# ecef_to_enu( x,y,z, latitude,longitude, xr,yr,zr )    Given Earth-centred, Earth-fixed coordinates of a point (x,y,z) and a reference point (xr,yr,zr),
#                                                       return a vector in East/North/Up coordinates representing the displacement between the two points (in meters).
#
#
# Altutude Azimuth coordinates are directions:
# - Altitude is height above the local horizon
# - N.B. NOTE Ambiguities in various definitions of azimuth angle
#   HiSPARC docs say measure from East and go anticlockwise  (See "Coordinate systems and units in HiSPARC"; Laat & Beigen, 2016)
#   Most astronomers start from North and go clockwise (i.e. round to the East first).
#
# When performing coordinate conversions, it is often much more convenient to work with unit vectors
# pointing to or from the direction to be transformed, and rotating these with matrix algebra
# before converting back to angular coordinates.
#
# Declination is the angle in the meridian plane containing the object measured N from the celestial equator.
# (It is the celestial equivalent of lattitude.)
#
# Hour Angle is the angle between two planes, 
# -  one containing the Zenith and the Earth's axis
# - and the other containing the object of interest and the Earth's axis.
# - It is measured as positive to the West.
#
# haDecToECEFUnitVec(ha,dec)                            Convert Hour Angle and Declination to a unit vector
#                                                       pointing along the direction of the shower track
#                                                       FROM the sky location designated by ha,dec.
#
# ECEFunitVecToHaDec(p,q,r)                             Take a unit vector in ECEF coordinates,
#                                                       pointing FROM a sky location and convert to Hour-  angle and Declination.
#
# ENUVecToAltAzm(p,q,r)                                 p q r and components of a vector in a right handed coordinate system
#                                                       in an EastNorthUp framework,
#                                                       that is, p and q correspond to x and y axes where z is x /\ y (i.e. the vector product of x and y).
#                                                       So, at the North Pole, x points South, y points East, and z points up.
#                                                       The vector is assumed to point INTO the origin, rather than OUT from the origin.
#
#
# Note that Hour Angle is the angle measured westwards from the meridian to the hour circle of the sky-point to be transformed.
# To convert Hour Angle to Right Ascension the local sidereal time (LST) is also required.
# 
#
# altAzmToHaDec(altitude,azimuth, latitude,longitude)   For a given latitude, longitude, rotate altitude,azimuth to hour angle and declination.
#
# haDecToAltAzm(ha,declination, latitude,longitude)     For a given latitude, longitude, rotate hour angle and declination to altitude,azimuth.
#
#------------------------------------------------------------------------------------------------------------------------------
# julianDay(year,month,day,hour)                        For a year, month, day and hour, return the Julian Day.
#
#------------------------------------------------------------------------------------------------------------------------------
# axisAngles( ha,dec, dbx,dby,dbz,dcx,dcy,dcz )         For a given hour angle and declination, and displacements of detectors B and C from detector A,
#                                                       return the angle between the shower direction and
#                                                       - the A-B line (thB)
#                                                       - the A-C line (thC)
#                                                       - the line perpendicular to the ABC plane (thC).
#                                                       The values are returned as a triplet list.
#
# stationStatToNumDict(stationNumbers)                  Given a list of station numbers, return a dictionary representing the mapping from
#                                                       station number to a character string of the form '/s<stationNumber>'
#
#------------------------------------------------------------------------------------------------------------------------------
# testRotate()                                          Set up a number of testcases, with a variety of angles, where the results are easy to calculate manually.
#                                                       we go in both directions, and should get back to the original inputs if all has worked correctly,
#                                                       using altAzToHaDec() and haDecToAltAz().
#
#==================================================================================================================================================================

import math         # Python standard library
from sapphire.api import API, Network, Station
from sapphire.time_util import GPSTime

from astropy   import coordinates as coord
from astropy   import units as u

uTestMode = False

clight   = 299792458.0
mtonlgt  = 1.0E9/clight             # Convert meters to nano-light seconds
radToDeg = 180.0/math.pi
degToRad = math.pi/180.0

#=======================================================================================================

def removeSpaces(string):
    """Remove all spaces from the input string, and return a new string"""
    newString = re.sub(' ','_',string)
    return newString
#=======================================================================================================

def combinationDictionary(stationNumList):
    """
        In:     a list of station numbers.
        Out:    a Dictionary and a List, 
                1) the keys are tuples containing every combination of stations with three or more members, 
                   (listed in ascending order) and the value is a unique combination (or configuration) id.
                2) The List contains the 
    """
#=======================================================================================================

    listSize = len(stationNumList)
    
    combToID = {}
    idToComb = []
    
    uniqueID = 0
    
    
    for cSize in range(3,listSize+1):
        cList = combinationList(cSize,listSize)
        for comb in cList:
            sList=[]
            for c in comb:
                sList.append(stationNumList[c])
                
            sList.sort()             # Ensure sorted in ascending order.
            tup_sList = tuple(sList)  # Create tuple from the list

            combToID[tup_sList]=uniqueID
            idToComb.append(tup_sList)
            
            uniqueID = uniqueID+1
            
    return (combToID,idToComb)


def combinationList(nsel,fromNum):
    """Produce a list that contains all combinations selecting nsel from fromNum.
    
        Output: a list of lists
            each element of the top level list is a list of nsel index numbers in range 1..fromNum.
    """
    
    complete = 0    
    count    = 0
    
    pmax = nsel-1  # The last index of the pick[] list.
    pick = [i for i in range(nsel)]      # The initial selection from the nmax stations. (Counting from 0.)

    plist = []
    plist.append(pick[:]) 

    
    free = pmax                 # The left-most index free to increment

    complete = False
    while(not complete):
        count = count+1
        
        for i in range(nsel-1):
            if(pick[i]< pick[i+1]-1):
                free = i
                break
        else:
            if(pick[pmax] < fromNum-1):
                free = pmax
            else:
                complete = True
                break
            
        pick[free]=pick[free]+1 
        for j in range(free):
            pick[j]=j
        plist.append(pick[:])
    
    return plist 
    
#============================================================================================================

def intraClusterDistance(dx,dy,dz):
    """ For a list of x,y,z coordinates (relative to a reference station)
        return the maximum distance between any two stations (in same units as dx,dy,dz).
    """
    maxDist = 0.0
    minDist = 1000000000.0
    sumDist = 0.0
    
    plist = combinationList(2,len(dx))
    for p in plist:
        (i1,i2) = p
        dist = distance(dx[p[0]],dx[p[1]], dy[p[0]],dy[p[1]], dz[p[0]],dz[p[1]])
        sumDist = sumDist + dist
        
        if dist > maxDist :
            maxDist = dist
        if dist < minDist:
            minDist = dist
            
        meanDist = sumDist/len(plist)
        
    return (maxDist,minDist,meanDist)
        
#============================================================================================================


def stationStatToNumDict(stationNumbers):
    """Given a list of station numbers,
        return a dictionary representing the mapping from
        station number to a character string of the form '/s<stationNumber>'"""

    stat_to_num    = {'/s%d' % u : u for u in stationNumbers}
    return stat_to_num


#============================================================================================================

def getStationsInCluster(cName):
    """For a given cluster identified by name,
        return a list of all the station numbers in that cluster."""
    
    allClusters = Network().clusters()
    
    for cluster in allClusters:
        if cName == cluster['name']:
            stationNumbers = Network().station_numbers(cluster=cluster['number'])

            return stationNumbers
   
def downloadClusterData(stationNums,start,end,toFile):

    station_groups = ['/s%d' % u for u in stationNums]

    print stationNums
    print station_groups
    
    cdata = tables.open_file(toFile, 'w')

    for station, group in zip(stationNums, station_groups):
        download_data(data, group, station, start, end)

#============================================================================================================
# Vector algebra

def includedAngle(vx,vy,vz, wx,wy,wz):
    """ Given two vectors: v,w
        defined by their components in three orthogonal directions: x,y,z
        work out the angle between them."""
    
    lenv = math.sqrt(vx*vx + vy*vy + vz*vz)
    lenw = math.sqrt(wx*wx + wy*wy + wz*wz)
    
    cosA = (vx*wx + vy*wy + vz*wz)/(lenv*lenw)
    A    = math.acos(max(-1.0,min(1.0,cosA)))

    return A

def vectorMagnitude(x,y,z):
    """ The absolute length of a vector."""
    return math.sqrt(x*x + y*y + z*z)

def distance(x1,y1,z1,x2,y2,z2):
    """Return distance between to point defined by x,y,z coordinates."""
    return vectorMagnitude(x1-x2,y1-y2,z1-z2)
    
def vectorProduct(x1,y1,z1, x2,y2,z2):
    """Return the vector product of two vectors, which can be used
       to find the direction at right angles to two other vectors."""

    x3 = y1*z2 - z1*y2
    y3 = z1*x2 - x1*z2
    z3 = x1*y2 - y1*x2

    return (x3,y3,z3)

def normaliseVector(x,y,z):
    """Returns a unit vector in the same direction as the input vector."""
    M = vectorMagnitude(x,y,z)

    if (M > 0.0):
        return (x/M,y/M,z/M)
    else:
        return (0.0,0.0,0.0)
        
#============================================================================================================
# Rotation matrices:
#
# A rotation about the Z axis by a positive angle a is represented by the matrix |x'| | cos(a)  sin(a)  0| |x|
#                                                                                |y'|=|-sin(a)  cos(a)  0| |y|
#                                                                                |z'| | 0       0       1| |z|
#
# A rotation about the X axis by a positive angle b is represented by the matrix |x'| | 1   0       0    | |x|
#                                                                                |y'|=| 0  cos(b)  sin(b)| |y|
#                                                                                |z'| | 0 -sin(b)  cos(b)| |z|
#

def rotateX(x,y,z,a):
    """Rotate an x,y,z vector around the X axis by right-handed angle a (radians)."""

    xp = x    
    yp = math.cos(a)*y + math.sin(a)*z
    zp =-math.sin(a)*y + math.cos(a)*z
    
    return (xp,yp,zp)  

def rotateY(x,y,z,a):
    """Rotate an x,y,z vector around the Y axis by right-handed angle a (radians)."""

    xp = math.cos(a)*x + math.sin(a)*z
    yp = y    
    zp =-math.sin(a)*x + math.cos(a)*z
    
    return (xp,yp,zp)
  
    
def rotateZ(x,y,z,a):
    """Rotate an x,y,z vector around the Z axis by right-handed angle a (radians)."""
    
    xp = math.cos(a)*x + math.sin(a)*y
    yp =-math.sin(a)*x + math.cos(a)*y
    zp = z
    
    return (xp,yp,zp)
    
#============================================================================================================
# Coordinate utilities
#
# Earth Centred, Earth Fixed coordinates: 
#   The origin is the centre of the Earth
#   put the Z axis along the Earth's rotation axis
#   put the X axis through the zero of lattitude and longitude (measuring longitude positive to the East).
#   put the Y axis through 90 degrees East, and zero lattitude
#


def gps_to_ecef(latitude,longitude,alt):
    """Convert GPS coordinates into Earth Centred, Earth Fixed coordinates.
       latitude, longitude in radians
       alt == altitude i meters.
       Return [x,y,z] in meters.
    """
    
    slong = math.sin(longitude)
    clong = math.cos(longitude)
    
    slat  = math.sin(latitude)
    clat  = math.cos(latitude)

    a = 6378137.0                      # Radius of the Earth in meters
    f = 1.0/298.257223563
    b = a*(1-f)
    e = math.sqrt( (2.0-f)*f )

    boa = (1-f)

    N = a/math.sqrt(1-e*e*slong*slong)

    x = (N + alt)*clat*clong
    y = (N + alt)*clat*slong
    z = ( N*(boa*boa) + alt )*slat
    
    return (x,y,z)          # Meters
    
#============================================================================================================
#
# ENU coordinates take the origin as a point on the Earth's surface
# - the X axis points along the meridian of longitude, to the North
# - the Y axis points along a line of lattitude, to the East
# - the Z axis point directly up to the zenith from the origin.

def ecef_to_enu( x,y,z, latitude,longitude, xr,yr,zr ):
    """Given Earth-centred, Earth-fixed coordinates of a point and a reference point,
        return a vector in East/North/Up coordinates representing the displacement between the two points,
        FROM the reference point xr,yr,zr TO x,y,x 

        Input:
            x,y,z                ECEF coordinates of a point
            latitude,longitude   Latitude and Longitude of the reference point
            xr,yr,zr             ECEF coordinates of a reference point."""
    
    xn = -math.sin(longitude)                   *(x-xr) + math.cos(longitude)                   *(y-yr)
    yn = -math.sin(latitude)*math.cos(longitude)*(x-xr) - math.sin(latitude)*math.sin(longitude)*(y-yr) + math.cos(latitude)*(z-zr)
    zn =  math.cos(latitude)*math.cos(longitude)*(x-xr) + math.cos(latitude)*math.sin(longitude)*(y-yr) + math.sin(latitude)*(z-zr)

    return (xn,yn,zn)


#============================================================================================================
#
# Convertion between (ECEF<--->ha,dec).

def haDecToECEFUnitVec(ha,dec):
    """Convert Hour Angle and Declination to an Earth Centred, Earth fixed unit vector
       pointing TO the sky location designated by ha,dec.
    
       Input:
          ha: hour angle (radians) measured positive westwards.
          dec: declination (radians) measured positive to the north from the equator.
    
       Output:
         [x,y,z] a direction vector with unit norm.
    
    """
    
    x =  math.cos(dec)*math.cos(ha)         # Note that Hour Angle is measured Westwards from the Meridian
    y = -math.cos(dec)*math.sin(ha)         # The negative sign arises because ECEF coord have positive Y to the East, while HA increases to the West.
    z =  math.sin(dec)
    
    return (x,y,z)


def ECEFUnitVecToHaDec(x,y,z):
    """Take a unit (NORMALISED) vector in ECEF coordinates,
       pointing TO a sky location
       and convert to Hour angle and Declination."""

    ha  = math.atan2( -y,x )
    dec = math.asin( min(max(z,-1.0),1.0) )    # Need to ensure that the argument is STRICTLY in the range (-1...1).

    return (ha,dec)
    
def ECEFVecToHaDec(x,y,z):
    """Take a vector in ECEF coordinates,
       pointing TO a sky location
       and convert to Hour angle and Declination."""
    
    norm = math.sqrt(x*x + y*y + z*z)

    ha  = math.atan2( -y,x )
    dec = math.asin( min(max(z/norm,-1.0),1.0) )    # Need to ensure that the argument is STRICTLY in the range (-1...1).

    return (ha,dec)

#------------------------------------------------------------------------------------------------
# Convertion between (ENU<--->alt,azm).

def AltAzmToENUunitVec(alt,azm):
    """e n u and components of a unit vector in a right handed coordinate system pointing to the sky in an East/North/Up framework. 
       That is, e and n correspond to x and y axes where z is x /\ y (i.e. the vector product of x and y).
       So, at the North Pole, x points South, y points East, and z points up.
       The vector is assumed to point TO the sky FROM the origin.
       """

    e = math.cos(alt)*math.sin(azm)
    n = math.cos(alt)*math.cos(azm)
    u = math.sin(alt)
    
    return (e,n,u)
    
    
def ENUunitVecToAltAzm(e,n,u):
    """e n u and components of a unit vector in a right handed coordinate system pointing to the sky in an East/North/Up framework. 
       That is, e and n correspond to x and y axes where z is x /\ y (i.e. the vector product of x and y).
       So, at the Equator, 0 longitude, x points East, y points North, and z points up.
       The vector is assumed to point TO the sky FROM the origin.
       Note that in this convention, azimuth is measured from the North round by East.
       """

    azm = math.atan2(e,n)                      # i.e zero when e=0 and n>0; pi/2 when e> 0 and n=0, pi when e=0 and n<0. 3pi/2 when e>0 and n=0.
    alt = math.asin( min(max(u,-1.0),1.0) )    # Need to ensure that the argument is STRICTLY in the range (-1...1).

    return (alt, azm)
   

#============================================================================================================
# Use of rotation matrices to inter-convert from (alt,azm)<-->(ha,dec) as a function of (lat,long)
# by first converting via a unit vector (in ENU or ECEF) and rotating this vector before going back to angles:
#                   (alt,azm <--> ENU) (ENU <-->ECEF) (ECEF <--> ha,dec)


def ENUtoECEFunitVec(e,n,u,latitude,longitude):
    """To convert from ENU to ECEF we need to 
       rotate around the ENU e axis, by (PI/2-latitude)
       and then around the ENU Z axis by -(p[i/2+longitude). 
    """

#    cosa = math.cos((math.pi/2.0+longitude))
#    sina = math.sin((math.pi/2.0+longitude))
#    cosb = math.cos(-(math.pi/2.0 - latitude))
#    sinb = math.sin(-(math.pi/2.0 - latitude))   
    
#    x =  cosa*e + sina*cosb*n + sina*sinb*u
#    y = -sina*e + cosa*cosb*n + cosa*sinb*u
#    z =   0.0*e -      sinb*n +      cosb*u     
#    return (x,y,z)

    (p,q,r) = rotateX(e,n,u,-(math.pi/2.0-latitude) )
    (x,y,z) = rotateZ(p,q,r,-(math.pi/2.0+longitude))
    
    return (x,y,z)


def ECEFtoENUunitVec(x,y,z,latitude,longitude):
    """To convert from ECEF to ENU we need to 
        1)  rotate around the ECEF z axis, by (pi/2+longitude)
        2)  and then around the ECEF x axis by -(PI/2-lattitude). 
    """

#    cosa = math.cos(math.pi/2.0+longitude)
#    sina = math.sin(math.pi/2.0+longitude)
#    cosb = math.cos((math.pi/2.0 - latitude))
#    sinb = math.sin((math.pi/2.0 - latitude))   
         
#    e =  cosa     *x + sina     *y
#    n = -sina*cosb*x + cosa*cosb*y + sinb*z
#    u =  sina*sinb*x - cosa*sinb*y + cosb*z

    (p,q,r) = rotateZ(x,y,z, (math.pi/2.0+longitude))
    (e,n,u) = rotateX(p,q,r, (math.pi/2.0-latitude) )

    return (e,n,u)

#------------------------------------------------------------------------------------------------
    
def altAzmToHaDec(altitude,azimuth, latitude,longitude):
    """For a given latitude, longitude, rotate altitude,azimuth to hour-angle and declination."""
    
    # Coordinates of a unit vector in the direction specified
    # using a right handed coordinate system for altitude,azimuth.

    (e,n,u)  = AltAzmToENUunitVec(altitude,azimuth)
    (x,y,z)  = ENUtoECEFunitVec(e,n,u,latitude,longitude)
    (ha,dec) = ECEFVecToHaDec(x,y,z)
    
    if(uTestMode):
        return (ha,dec,e,n,u,x,y,z)
    else:
        return (ha,dec)


def haDecToAltAzm(ha,declination, latitude,longitude):
    """For a given latitude, longitude, rotate hour angle and declination to altitude,azimuth."""

    # Coordinates of a unit vector in the direction specified,
    # using a right handed coordinate system.

    (x,y,z)            = haDecToECEFUnitVec(ha,declination)
    (e,n,u)            = ECEFtoENUunitVec(x,y,z,latitude,longitude)    
    (altitude,azimuth) = ENUunitVecToAltAzm( e,n,u )

    if(uTestMode):
        return (altitude,azimuth,e,n,u,x,y,z)
    else:
        return (altitude,azimuth)


#============================================================================================================
# Conversion to Galactic Coordinates, from ra,dec

def raDecToGalactic(ra_rad,dec_rad):
    """ Right Ascension and declination given in radians.
        Returns (l,b) in radians
    """
    c = coord.SkyCoord(ra=ra_rad*u.rad,dec=dec_rad*u.rad)
    l = c.galactic.l.rad
    b = c.galactic.b.rad
    return (l,b)

def hammerAitoff(long_rad,lat_rad):
    x = math.sqrt(8)*math.cos(lat_rad)*math.sin(long_rad/2.0)/math.sqrt(1+math.cos(lat_rad)*math.cos(long_rad/2.0))
    y = math.sqrt(2)*math.sin(lat_rad)                       /math.sqrt(1+math.cos(lat_rad)*math.cos(long_rad/2.0))
    return(x,y)
    
#============================================================================================================

def julianDay(year,month,day,hour):
    """For a year, month, day and hour, return the Julian Day."""
    
    a = int( (14 - month)/12 )
    A = int((year - a)/100)
    B= 2 -A + int(A/4)
    C = int(365.25*(year - a))
    D = int(30.60001*(month+ 12*a + 1))
    JD = B + C + D + day + hour/24 + 1720994.5

    return JD

def GPSTimeToDateAndTime(gpstime):
    """Return string for date and time"""
    dtime = GPSTime(gpstime)    # Create a GPS datetime object 
    return dtime.description()
    
#============================================================================================================

def axisAngles( ha,dec, dbx,dby,dbz,dcx,dcy,dcz ):
    """For a given hour angle and declination,
    and displacements of detectors B and C from detector A,
    return the angle between the shower direction and
    - the A-B line (thB)
    - the A-C line (thC)
    - the line perpendicular to the ABC plane (thC)."""

    (p,q,r) = haDecToECEFUnitVec(ha,dec)
    
    (ddx,ddy,ddz) = vectorProduct( dbx,dby,dbz, dcx,dcy,dcz )
    
    thD = includedAngle(p,q,r, ddx,ddy,ddz)
    thB = includedAngle(p,q,r, dbx,dby,dbz)
    thC = includedAngle(p,q,r, dcx,dcy,dcz)


    return(thB,thC,thD)


#============================================================================================================

def testENUAltAzm():

    """Set up a number of testcases, with a variety of angles, where the results are easy to calculate manually.
       we go in both directions, and should get back to the original inputs if all has worked correctly."""
    
    altitudes = [-89.999, -45.0, 0.001,   45.0,  89.999]
    azimuths  = [-180.0,-135.0, -90.0, -45.0, 0.001,   45.0,  90.0, 135.0, 180.0]
    

    print "\nTest alt,azm <---> ENU:"
        
    print "   alt1     azm1  -->     e1       n1       u1   -->    alt2     azm2  -->     e2       n2       u2 ",
    print " |   dalt    dazm     de      dn      du "
    for alt1 in altitudes:
        for azm1 in azimuths:

            altrad = alt1*degToRad
            azmrad = azm1*degToRad

            (e1,n1,u1)  = AltAzmToENUunitVec(altrad,azmrad)    
            (alt2,azm2) = ENUunitVecToAltAzm(e1,n1,u1)
            (e2,n2,u2)  = AltAzmToENUunitVec(alt2,azm2)   
         
            
            print "%8.3f %8.3f --> %8.3f %8.3f %8.3f --> %8.3f %8.3f --> %8.3f %8.3f %8.3f"%(alt1,azm1,e1,n1,u1,radToDeg*alt2,radToDeg*azm2,e2,n2,u2),
            print "| %7.3f %7.3f  %7.3f %7.3f %7.3f"%(alt1-radToDeg*alt2,azm1-radToDeg*azm2,e1-e2,n1-n2,u1-u2)
    

def testECEFHaDec():
    
    """ Set up a number of testcases, with a variety of angles, where the results are easy to calculate manually.
        we go in both directions, and should get back to the original inputs if all has worked correctly.
        In this case, we start with a conversion
            Hour-angle, declination ---> ECEF unit vector (points to a sky position)
            ECEF Vector             ---> HA2,Dec2
            HA2,Dec2                ---> ECEF unit vector (2)
    """
    
    hourAngles    = [-180.0,-135.0, -90.0 -45.0, 0.001,   45.0,  90.0, 135.0, 180.0]
    declinations  = [-89.999, -45.0, 0.001,   45.0,  89.999]

    print "\nTest Ha,Dec <---> ECEF:"
        
    print "   ha1      dec1  -->     x1       y1       z1   -->   ha2      dec2  -->      x2       y2       z2 ",
    print " |   dha      ddec     dx      dy      dz"

    for ha1 in hourAngles:
        for dec1 in declinations:

            harad   = ha1*degToRad
            decrad = dec1*degToRad

            (x1,y1,z1)  = haDecToECEFUnitVec(harad,decrad)    
            (ha2,dec2)  = ECEFVecToHaDec(x1,y1,z1)
            (x2,y2,z2)  = haDecToECEFUnitVec(ha2,dec2)   
         
            
            print "%8.3f %8.3f --> %8.3f %8.3f %8.3f --> %8.3f %8.3f --> %8.3f %8.3f %8.3f"%(ha1,dec1,x1,y1,z1,radToDeg*ha2,radToDeg*dec2,x2,y2,z2),
            print "| %7.3f %7.3f  %7.3f %7.3f %7.3f"%(ha1-radToDeg*ha2,dec1-radToDeg*dec2,x1-x2,y1-y2,z1-z2)

def testECEF_ENU():

    """Set up a number of testcases, with a variety of vectors, where the results are easy to calculate manually.
       we go in both directions, and should get back to the original inputs if all has worked correctly."""

    print "\nTest ECEF <==> ENU:"
    
    xs=[-1.0, 0.0, 1.0]
    ys=[-1.0, 0.0, 1.0]
    zs=[-1.0, 0.0, 1.0]
    
#    latitudes  = [-89.999, -45.0, 0.0, 45.0, 89.999]
#    longitudes = [-179.999, -135.0, -90, -45.0, 0.0, 45.0, 90.0, 135.0, 180.0]    
    latitudes  = [-90.0, 0.0, 90.0]
    longitudes = [-90.0, 0.0, 90.0,]   
    print "  lat      long   |    x1       y1       z1            e1       n1       u1            x2       y2       z2            e2       n2       u2   ",
    print "|    dx     dy     dz     |   de     dn     du" 

    for lat in latitudes:
        for long in longitudes:
            print " "
            for x in xs:
                for y in ys:
                    for z in zs:
                        
                        if(abs(x) + abs(y) + abs(z) == 1.0):
                        
                            (x1,y1,z1)=normaliseVector(x,y,z)
                        
                            (e1,n1,u1) = ECEFtoENUunitVec(x1,y1,z1,degToRad*lat,degToRad*long)
                            (x2,y2,z2) = ENUtoECEFunitVec(e1,n1,u1,degToRad*lat,degToRad*long)
                            (e2,n2,u2) = ECEFtoENUunitVec(x2,y2,z2,degToRad*lat,degToRad*long)
                
                            print "%8.3f %8.3f | %8.3f %8.3f %8.3f ---> %8.3f %8.3f %8.3f ---> %8.3f %8.3f %8.3f ---> %8.3f %8.3f %8.3f"%(lat,long,x1,y1,z1,e1,n1,u1,x2,y2,z2,e2,n2,u2),
                            print "| %7.3f %7.3f %7.3f | %7.3f %7.3f %7.3f"%(x1-x2,y1-y2,z1-z2,e1-e2,n1-n2,u1-u2)

def testAltAzmHaDec():

    """Set up a number of testcases, with a variety of angles, where the results are easy to calculate manually.
       we go in both directions, and should get back to the original inputs if all has worked correctly."""

    latitudes = [ -89.999,-60.0,  -45.0, -30.0,  0.001, 30.0,   45.0, 60.0, 89.999]
    altitudes = [ -89.999,-60.0,  -45.0, -30.0,  0.001, 30.0,   45.0, 60.0, 89.999]

    longitudes= [-179.999,  -89.999,-60.0, -45.0, -30.0, 0.001, 30.0, 45.0, 60.0, 89.999, 179.999]
    azimuths  = [-179.999,  -89.999,-60.0, -45.0, -30.0, 0.001, 30.0, 45.0, 60.0, 89.999, 179.999]    


#    latitudes = [          -89.999, 0.0001, 89.999         ]
#    longitudes= [ -179.999,-89.999, 0.0001, 89.999, 179.999]

#    altitudes = [                    0.0001, 89.999         ]
#    azimuths  = [ -179.999, -89.999, 0.0001, 89.999, 179.999] 

    print "\nTest Alt,Azm <---> Ha,Dec"    
    print "    lat    long |",
    print "     alt     azm -->",
    print "     e1     n1     u1 -->",
    print "     X1     Y1     Z1 -->",
    print "      ha     dec -->",
    print "     X2     Y2     Z2 -->",
    print "     e2     n2     u2 -->",    
    print "altn   error      azmn   error     |",
    print "   dx1    dx2    dx2   |",
    print "   de2    dn2    du2 "

    for latitude in latitudes:
        latrad = latitude*degToRad

        for longitude in longitudes:
            longrad = longitude*degToRad

            print " "
            
            for altitude in altitudes:
                for azimuth in azimuths:

                    altrad = altitude*degToRad
                    azmrad = azimuth *degToRad
                    (ha,dec,            e1,n1,u1, x1,y1,z1) = altAzmToHaDec(altrad,azmrad,  latrad,longrad)    
                    (altituden,azimuthn,e2,n2,u2, x2,y2,z2) = haDecToAltAzm(ha,    dec,     latrad,longrad)

                    hadeg = ha       *radToDeg
                    decdeg= dec      *radToDeg
                    altdeg= altituden*radToDeg
                    azmdeg= azimuthn *radToDeg

                    if( abs(e1-e2) > 1.E-4 or abs(n1-n2) > 1.E-4 or abs(u1-u2) > 1.E-4):
                        flag = "Error"
                    else:
                        flag = " "

                    if( abs(x1-x2) > 1.E-4 or abs(y1-y2) > 1.E-4 or abs(z1-z2) > 1.E-4):
                        flagV = "*"
                    else:
                        flagV = " "

                    print "%7.2f %7.2f |"    % (latitude,longitude),
                    print " %7.2f %7.2f -->" % (altitude,azimuth),
                    print " %6.3f %6.3f %6.3f -->" % (e1,n1,u1),
                    print " %6.3f %6.3f %6.3f -->" % (x1,y1,z1),
                    print " %7.2f %7.2f -->" % (hadeg,decdeg),
                    print " %6.3f %6.3f %6.3f -->" % (x2,y2,z2),
                    print " %6.3f %6.3f %6.3f"    % (e2,n2,u2),
                    print " %7.2f (%7.2f) %7.2f (%7.2f) " % (altdeg,altdeg-altitude,azmdeg,azmdeg-azimuth),
                    print " | %6.3f %6.3f %6.3f " % ((x1-x2),(y1-y2),(z1-z2)),
                    print " | %6.3f %6.3f %6.3f %s%s"%((e1-e2),(n1-n2),(u1-u2),flagV,flag)


    
    
def testCluster():
  
  allClusters = Network().clusters()
  
  for cluster in allClusters:
          stationNumbers = Network().station_numbers(cluster=cluster['number'])
          print cluster, stationNumbers

                           
if __name__ == "__main__":
    import sys,re,math

    uTestMode = True

#    testCluster()
    testENUAltAzm() 
    testECEFHaDec()  
    testECEF_ENU()
    testAltAzmHaDec()
        
